/*
**	branch_inst.c	- Branch analyzer instrumentation code
**
**		This is the instrumentation code for the branch analyzer.  This
**		Atck tool is a simple example that prints statistics about the
**		branch instructions in an application.
**
**		Usage:
**			branch <app> [<report_proc>]
**
**		Where:
**			<app>          - The application to analyze
**			<report_proc>  - Procedure at which data is reported
**
**		If the <report_proc> procedure is specified, the branch tool reports
**		branch statistic data each time that procedure is executed.  In
**		addition, the user can press <ENTER> on the keyboard to report the
**		branch data.  The user can press x-<ENTER> (the 'x' key and then ENTER)
**		to stop the application and terminate the branch tool.
**
**		The branch statistic data shows how many conditional branch
**		instructions have been executed and what percentage have been taken.
*/

#include <stdio.h>
#include <ctype.h>
#include <atck.h>
#include <atckps2.h>
#include "branch.h"


static atck_iprog_t *	Instrument(atck_sesn_t *, atck_prog_t *, const char *);
static atck_bool_t   	RunIt(atck_sesn_t *, atck_cfg_t *, atck_iprog_t *, atck_prog_t*);
static void          	HandleEvt(atck_tevt_t *, void *, void *);
static void          	PrintData(atck_run_t *, atck_addr_t, atck_prog_t*);

static atck_uint32_t	IsBranchLikely(atck_inst_t* inst);
static void				AnalyzeBranch(atck_uint32_t branchID, atck_inst_t* inst);
brdat_t* pBuffer;
atck_uint32_t nBranches;


int main(
	int    	argc,
	char **	argv)
{
	atck_sesn_t * 	psesn;
	atck_cfg_t *  	pcfg;
	atck_prog_t * 	pprog;
	atck_iprog_t *	piprog;
	char *        	pnam_app;
	char *        	pnam_proc;


	/*
	 * Get the command line arguments.
	 */
	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Usage: branch <app> [<procedure>]\n");
		return(1);
	}

	pnam_app = argv[1];
	if (argc == 3)
		pnam_proc = argv[2];
	else
		pnam_proc = 0;

	/*
	 * Open the input application and our own tool's analysis file.
	 */
	psesn = atck_session("1.0", ATCK_FLAGS_NONE);
	pcfg = atck_config_new(psesn, "<syscfg>");
	pprog = atck_open(psesn, pnam_app, "<tooldir>branch.anal", pcfg, 0);
	if (!pprog) {
		atck_endsession(psesn);
		return(1);
	}

	/*
	 * Instrument the application.
	 */
	piprog = Instrument(psesn, pprog, pnam_proc);
	if (!piprog) {
		atck_endsession(psesn);
		return(1);
	}

	/*
	 * Run the instrumented application, and print the branch data.
	 */
	if (!RunIt(psesn, pcfg, piprog, pprog)) {
		atck_endsession(psesn);
		return(1);
	}

	atck_endsession(psesn);
	return(0);
}


/*
**  Instrument the application and return a handle to the instrumented
**  program.  Return NULL on error.
*/
static atck_iprog_t *Instrument(
	atck_sesn_t *	psesn,
	atck_prog_t *	pprog,
	const char * 	pnam_proc)
{
	atck_proto_t * 	pproto;
	atck_imgit_t * 	pimgi;
	atck_img_t *   	pimg;
	atck_procit_t *	ppi;
	atck_proc_t *  	pproc;
	atck_ent_t *   	pent;
	atck_bbit_t *  	pbi;
	atck_bb_t *    	pb;
	atck_instit_t *	pii;
	atck_inst_t *  	pi;
	atck_iprog_t * 	piprog;
	
	atck_uint32_t	branchID = 0;


	/*
	 * If a procedure was specified, add instrumentation at that procedure to
	 * call the Report() analysis routine.  The Report() routine will cause
	 * the branch data to be reported.
	 */
	if (pnam_proc) {
		pent = atck_ent_byname(pprog, pnam_proc);
		if (pent) {
			pproto = atck_analproto(pprog, "void Report()");
			atck_ent_callbefore(pent, pproto);
		}
		else {
			fprintf(stderr, "Warning: procedure '%s' does not exist.\n",
				pnam_proc);
		}
	}

	/*
	 * This is the main part of the instrumentation.  Scan through every
	 * instruction in the application looking for conditional branches.  Add
	 * instrumentation at each to call the TraceBr() analysis routine.
	 */
	pproto = atck_analproto(pprog, "void TraceBr(uint32, uint32, valuint32)");

	pimgi = atck_imgit_new(pprog);
	while (pimg = atck_imgit_next(pimgi)) {
		ppi = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(ppi)) {
			if (atck_proc_attr(pproc) == ATCK_ATTR_INSTR) {
				pbi = atck_proc_bbit_new(pproc);
				while (pb = atck_bbit_next(pbi)) {
					pii = atck_bb_instit_new(pb, ATCK_FLAGS_ITLIFE);
					while (pi = atck_instit_next(pii)) {
						if (atck_inst_iscbranch(pi) &&
							atck_inst_isallowed(pi, ATCK_TAKEN))
						{
							atck_inst_callbefore(pi, pproto, branchID, IsBranchLikely(pi), ATCK_TAKEN);
							branchID++;
						}
					}
					atck_instit_free(pii);
				}
				atck_bbit_free(pbi);
			}
		}
		atck_procit_free(ppi);

		/*
		 * Write out an instrumented version of this image.
		 */
		atck_img_write(pimg);
	}
	atck_imgit_free(pimgi);
	
	nBranches = branchID;
	pBuffer = atck_calloc(psesn, nBranches, sizeof(*pBuffer));
	atck_analtype(psesn, "BrDat" BRDAT_ANALTYPE);
	pproto = atck_analproto(pprog, "void Initialize(BrDat[])");
	atck_callbefore(pprog, pproto, nBranches, pBuffer);

	/*
	 * Finish writing out the instrumented application.
	 */
	piprog = atck_finish_write(pprog);
	return(piprog);
}


/*
**  Run the instrumented program and print the branch statistics when
**  requested.  Return TRUE on success, FALSE on failure.
*/
static atck_bool_t RunIt(
	atck_sesn_t * 	psesn,
	atck_cfg_t *  	pcfg,
	atck_iprog_t *	piprog,
	atck_prog_t*	pprog)
{
	atck_dev_t * 	pdev;
	atck_run_t * 	prun;
	atck_status_t	stat;
	atck_addr_t  	addr_brdat;
	char         	buf[80];


	/*
	 * Get the target address of the BranchData variable in the analysis
	 * code.  We will need to read the branch statistic data from this
	 * address below.
	 */
	addr_brdat = atck_anal_symaddr(piprog, "BranchData");
	if (!addr_brdat) {
		fprintf(stderr, "Unable to find symbol 'BranchData'.\n");
		return(ATCK_FALSE);
	}

	/*
	 * Connect to the target system.  Note, 'pcfg' tells us which system to
	 * connect to.  Events from the target system are handled by HandleEvt()
	 * below.
	 */
	pdev = atck_connect(psesn, pcfg, HandleEvt, &pprog);
	if (!pdev)
		return(ATCK_FALSE);

	/*
	 * Download the instrumented application and start it running.
	 */
	prun = atck_idownload(pdev, piprog, 0, &addr_brdat);
	atck_continue(prun);

	/*
	 * Stop the application whenever the user presses ENTER.  The application
	 * also stops if it executes the Report() analysis routine.  Either way,
	 * a STOPPED event is sent to the handle_evt() handler below, which causes
	 * the branch statistic data to be printed.
	 */
	printf("Press ENTER to get data, x-ENTER to quit\n");
	for (;;) {
		fgets(buf, sizeof(buf), stdin);
		stat = atck_status(prun);
		if (toupper(buf[0]) == 'X' || stat == ATCK_STATUS_EXITED ||
			stat == ATCK_STATUS_FAULTED || stat == ATCK_STATUS_ERROR)
		{
			break;
		}
		atck_stop(prun);
	}

	atck_disconnect(pdev);
	return(ATCK_TRUE);
}


/*
**  The Atck framework calls this routine whenever there is an event from
**  the target system.
*/
static void HandleEvt(
	atck_tevt_t *	pevt,
	void *       	ppprog,
	void *       	ppbuffer)
{
	atck_addr_t*	paddr = ppbuffer;
	atck_prog_t*	pprog = *((atck_prog_t**)ppprog);
	atck_addr_t  	addr_brdat;
	atck_run_t * 	prun;
	atck_status_t	stat;
	
	switch (atck_tevt_type(pevt))
	{
	case ATCK_TEVT_STOPPED:
		/*
		 * The application has stopped, either because the user pressed ENTER,
		 * because it executed the Report() analysis routine, or because it
		 * has terminated.  Regardless, we print the branch statistic data.
		 */
		prun = atck_tevt_run(pevt);
		addr_brdat = *paddr;
		PrintData(prun, addr_brdat, pprog);

		/*
		 * Let the user know if the application terminated.
		 */
		stat = atck_status(prun);
		if (stat == ATCK_STATUS_EXITED || stat == ATCK_STATUS_FAULTED)
			printf("\nApplication terminated, press ENTER to quit.\n");
		break;

	case ATCK_TEVT_PRINT:
		/*
		 * The target wants to print some data to the host.  We choose to
		 * ignore this to avoid cluttering the output.  However, the print
		 * string can be displayed by uncommenting the line below.
		 *
		 * printf("%s", atck_tevt_printstr(pevt));
		 */
		break;

	case ATCK_TEVT_LOADPCT:
		/*
		 * These events show the progress of the download operation.
		 */
		printf("Downloading... %u%%\n", atck_tevt_loadpct(pevt));
		break;
	}
}


/*
**  Upload the branch statistic data from the instrumented application,
**  print it out, and continue the application.
*/
static void PrintData(
	atck_run_t *	prun,
	atck_addr_t 	addr_brdat,
	atck_prog_t*	pprog)
{
	atck_imgit_t * 	pimgi;
	atck_img_t *   	pimg;
	atck_procit_t *	ppi;
	atck_proc_t *  	pproc;
	atck_bbit_t *  	pbi;
	atck_bb_t *    	pb;
	atck_instit_t *	pii;
	atck_inst_t *  	pi;
	atck_uint32_t	branchID = 0;
	atck_addr_t		addr_buffer;

	/*
	 * Read the contents of the BranchData pointer in the analysis code.
	 */
	atck_readdata(prun, addr_brdat, &addr_buffer, "addr", 1);
	
	/*
	 * Read the contents of the BranchData[] buffer in the analysis code.
	 */
	atck_readdata(prun, addr_buffer, pBuffer, "BrDat", nBranches);

	pimgi = atck_imgit_new(pprog);
	while (pimg = atck_imgit_next(pimgi))
	{
		ppi = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(ppi))
		{
			if (atck_proc_attr(pproc) == ATCK_ATTR_INSTR)
			{
				pbi = atck_proc_bbit_new(pproc);
				while (pb = atck_bbit_next(pbi))
				{
					pii = atck_bb_instit_new(pb, ATCK_FLAGS_ITLIFE);
					while (pi = atck_instit_next(pii))
					{
						if (atck_inst_iscbranch(pi) &&
							atck_inst_isallowed(pi, ATCK_TAKEN))
						{
							AnalyzeBranch(branchID, pi);
							branchID++;
						}
					}
					atck_instit_free(pii);
				}
				atck_bbit_free(pbi);
			}
		}
		atck_procit_free(ppi);

		/*
		 * Write out an instrumented version of this image.
		 */
		atck_img_release(pimg);
	}
	atck_imgit_free(pimgi);

	/*
	 * Zero out the BranchData structure in the analysis code to reset
	 * the branch statistic data.
	 */
	while (branchID > 0)
	{
		pBuffer[branchID - 1].nbranches = 0;
		pBuffer[branchID - 1].ntaken = 0;
		--branchID;
	}
	atck_writedata(prun, addr_buffer, pBuffer, "BrDat", nBranches);

	/*
	 * Start the application running again.
	 */
	atck_continue(prun);
}

/*
**  Determine whether the specified branch instruction is likely
**	to branch or not.
*/
static atck_uint32_t IsBranchLikely(atck_inst_t* inst)
{
	/* An unconditional branch is, of course, VERY likely! */
	if (atck_inst_isubranch(inst))
		return 1;
		
	switch (atck_inst_op(inst))
	{
		case ATCKMIPS_OP_BEQL:
		case ATCKMIPS_OP_BGEZALL:
		case ATCKMIPS_OP_BGEZL:
		case ATCKMIPS_OP_BGTZL:
		case ATCKMIPS_OP_BLEZL:
		case ATCKMIPS_OP_BLTZL:
		case ATCKMIPS_OP_BNEL:
		case ATCKMIPS_OP_BC1FL:
		case ATCKMIPS_OP_BC1TL:
		case ATCKPS2_OP_BC2FL:
		case ATCKPS2_OP_BC2TL:
		case ATCKPS2_OP_BC0FL:
		case ATCKPS2_OP_BC0TL:
			return 1;
			
		default:
			/* FALL THROUGH... */
			;
	}		
	
	return 0;
}

static void	AnalyzeBranch(atck_uint32_t branchID, atck_inst_t* inst)
{
	atck_uint32_t pct;
	const char* name;
	unsigned line;
	
	if (pBuffer[branchID].nbranches)
	{
		pct = (pBuffer[branchID].ntaken * 100) / pBuffer[branchID].nbranches;
		if (pct < 50)
		{
			name = atck_inst_file(inst);
			if (name != NULL)
			{
				/* No sense reporting the branch if we can't tell the name
				** of the file: What would the user do with the info?
				*/
				
				line = atck_inst_line(inst);
			
				printf("File %s, line %d: ", name, line);
				
				if (IsBranchLikely(inst))
					printf("Branch-likely");
				else
					printf("Branch-unlikely");
					
				printf(" accurate only %d%% of the time (%d out of %d branches)\n", pct, pBuffer[branchID].ntaken, pBuffer[branchID].nbranches);
			}
		}			
	}
}
