#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>

#include <atck.h>
#include <atcktypes.h>

#include "RunContext.h"

using namespace std;

atck_sesn_t* RunContext::s_pSession = NULL;
atck_uint32_t RunContext::s_nSessionLock = 0;
atck_cfg_t* RunContext::s_pConfig = NULL;
atck_dev_t* RunContext::s_pDev = NULL;
atck_uint32_t RunContext::s_nDevLock = 0;

bool RunContext::InitSession(const char* app)
{
	if (s_pSession == NULL)
	{
		s_pSession = atck_session("1.0", ATCK_FLAGS_NONE);
		s_nSessionLock++;
	}
	
	if (!s_pSession)
		return false;
		
	if (s_pConfig == NULL)
		s_pConfig = atck_config_new(s_pSession, "<syscfg>");
		
	m_pProg = atck_open(s_pSession, app, "<tooldir>anal_code.elf", s_pConfig, "<dir><base>_inliner<suf>");
	if (!m_pProg)
		return false;
		
	return true;
}

RunContext::~RunContext()
{
	// Clear lists
	for (CandidateIterator candit = m_Candidates.begin(); candit != m_Candidates.end(); candit++)
		delete *candit;
		
	for (CallSiteIterator callit = m_CallSites.begin(); callit != m_CallSites.end(); callit++)
		delete *callit;
		
	if (m_pCounters)
		atck_free(s_pSession, m_pCounters);
		
	if (m_pProg)
		atck_close(m_pProg);
		
	if (m_pIProg)
		atck_iclose(m_pIProg);

	UnlockDevice();
	
	if (s_pSession)
	{
		--s_nSessionLock;
		if (!s_nSessionLock)
		{
			atck_endsession(s_pSession);
			s_pSession = NULL;
		}
	}
}

bool RunContext::BuildCandidateList(void)
{
	atck_imgit_t* 	pimgit;
	atck_img_t*   	pimg;
	atck_procit_t*	pprocit;
	atck_proc_t*  	pproc;
	atck_uint32_t	imgID;
	
	imgID = 0;
	pimgit = atck_imgit_new(m_pProg);
	while (pimg = atck_imgit_next(pimgit))
	{
		pprocit = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(pprocit))
		{
			Candidate* c = new Candidate(pproc, imgID);
			if (c->IsCandidate())
				m_Candidates.push_back(c);
			else
				delete c;
		}
		atck_procit_free(pprocit);

		atck_img_release(pimg);
		++imgID;
	}
	atck_imgit_free(pimgit);
	
	return true;
}

Candidate* RunContext::FindTargetInList(atck_addr_t addr, atck_uint32_t imgID)
{
	if (!addr)
		return NULL;
		
	// Step through CandidateList, checking each for match
	for (CandidateIterator it = m_Candidates.begin(); it != m_Candidates.end(); it++)
	{
		if ((*it)->DoesAddressMatch(addr, imgID))
			return *it;
	}
	
	return NULL;
}

bool RunContext::Instrument(void)
{
	atck_imgit_t* 	pimgit;
	atck_img_t*   	pimg;
	atck_uint32_t	imgID;
	atck_procit_t*	pprocit;
	atck_proc_t*  	pproc;
	atck_callit_t*	pcallit;
	atck_call_t*	pcall;
	atck_proto_t*	pproto;
	Candidate*		ptarget;
	
	pproto = atck_analproto(m_pProg, "void LogCall(uint32)");
	
	imgID = 0;
	m_nCounters = 0;
	pimgit = atck_imgit_new(m_pProg);
	while (pimg = atck_imgit_next(pimgit))
	{
		pprocit = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(pprocit))
		{
			pcallit = atck_proc_callit_new(pproc);
			while (pcall = atck_callit_next(pcallit))
			{
				ptarget = FindTargetInList(atck_call_targattr(pcall), imgID);
				if (ptarget)
				{
					m_CallSites.push_back(new CallSite(atck_img_inst_byaddr(pimg, atck_call_addr(pcall)), ptarget));
					atck_call_callbefore(pcall, pproto, m_nCounters);
					++m_nCounters;
				}
			}
		}
		atck_procit_free(pprocit);

		atck_img_write(pimg);
		++imgID;
	}
	atck_imgit_free(pimgit);

	m_pCounters = (atck_uint32_t*)atck_calloc(s_pSession, m_nCounters, sizeof(*m_pCounters));
	pproto = atck_analproto(m_pProg, "void Initialize(uint32[])");
	atck_callbefore(m_pProg, pproto, m_nCounters, m_pCounters);
	
	m_pIProg = atck_finish_write(m_pProg);
	
	return true;
}

bool RunContext::Run(const char* prompt)
{
	char        	buf[80];

	if (!LockDevice())
		return false;

	m_pRun = atck_idownload(s_pDev, m_pIProg, 0, this);
	atck_continue(m_pRun);

	printf("%s", prompt);
	fgets(buf, sizeof(buf), stdin);
	atck_stop(m_pRun);

	return true;
}

void RunContext::HandleStop(void)
{
	if (!m_aTargetBuffer)
	{
		atck_addr_t ptr = atck_anal_symaddr(m_pIProg, "pCounters");
		if (ptr)
			atck_readdata(m_pRun, ptr, &m_aTargetBuffer, "addr", 1);
			
		if (!m_aTargetBuffer)
			printf("Unable to find analysis symbols.\n");
	}
	
	atck_readdata(m_pRun, m_aTargetBuffer, m_pCounters, "uint32", m_nCounters);
	
	// Copy results into the call site list
	for (int i = 0; i < m_nCounters; i++)
		m_CallSites[i]->SetCallCount(m_pCounters[i]);
		
	// Now sort call sites by target address
	sort(m_CallSites.begin(), m_CallSites.end(), CompCallSites());
	
	// Now, for each candidate, print out results
	for (CandidateIterator candit = m_Candidates.begin(); candit != m_Candidates.end(); candit++)
	{
		// 	Print out candidate info
		(*candit)->Display();
		
		CallSiteIterator callit = m_CallSites.begin();
		while (callit != m_CallSites.end() && !((*callit)->DoesCandidateMatch(*candit)))
			callit++;
			
		while (callit != m_CallSites.end() && (*callit)->DoesCandidateMatch(*candit))
		{
			(*callit)->Display();
			delete *callit;
			callit = m_CallSites.erase(callit);
		}
	}
}

//==========================================
static void HandleEvt(atck_tevt_t* pevt, void* pdevdata, void* prundata)
{
	RunContext*		pctx = (RunContext*)prundata;
	
	switch (atck_tevt_type(pevt))
	{
		case ATCK_TEVT_STOPPED:
			pctx->HandleStop();
			break;

		case ATCK_TEVT_PRINT:
			break;

		case ATCK_TEVT_LOADPCT:
			printf("Downloading... %d%%\n", atck_tevt_loadpct(pevt));
			break;
	}
}

void RunContext::UnlockDevice(void)
{
	if (s_pDev)
	{
		s_nDevLock--;
		if (!s_nDevLock)
		{
			atck_disconnect(s_pDev);
			s_pDev = NULL;
		}
	}
}

bool RunContext::LockDevice(void)
{
	if (!s_pDev)
	{
		s_pDev = atck_connect(s_pSession, s_pConfig, HandleEvt, NULL);
		if (!s_pDev)
			return false;
	}
	
	s_nDevLock++;
	return true;
}
