#include <atck.h>
#include <atckps2.h>
#include <stdio.h>

/*
**	barebones.c	- The world's least-useful ATtaCK tool
**
**		This tool counts the number of procedures an
**		application executes. That's all. Output goes to stdout.
**
**		Usage:
**			barebones [-r] <app>
**
**		Where:
**			-r		  - Indicates that the app should be downloaded and
**					    run on the target
**			<app>     - The application to "analyze"
*/

static void HandleEvt(atck_tevt_t*, void*, void*);

atck_addr_t	pcounter;

int main(int argc, char** argv)
{
	atck_sesn_t*	psesn;
	atck_cfg_t* 	pcfg;
	atck_prog_t*	pprog;
	atck_imgit_t* 	pimgit;
	atck_procit_t*	pprocit;
	atck_img_t*   	pimg;
	atck_proc_t*  	pproc;
	atck_proto_t* 	pinitproto;
	atck_proto_t*	pcountproto;
	atck_iprog_t* 	piprog;
	atck_dev_t*		pdev;
	atck_run_t*		prun;

	char* 			szapp;
	int				brun = 0;
	int				nprocs = 0;

	/*
	 * 1. Read in user options
	 */
	if (argc < 2)
	{
		fprintf(stderr, "Usage: barebones [-r] <app>\n");
		fprintf(stderr, "-r indicates to download and run app on target\n");
		return(1);
	}
	else
	{
		if (argv[1][0] == '-')
		{
			brun = 1;
			szapp = argv[2];
		}
		else
			szapp = argv[1];
	}

	/*
	 * 2. Initialize an ATtaCK session
	 */
	psesn = atck_session("1.0", ATCK_FLAGS_NONE);

	/*
	 * 3. Read in system options
	 */
	pcfg = atck_config_new(psesn, "<syscfg>");

	/*
	 * 4. Open application
	 */
	pprog = atck_open(psesn, szapp, "<tooldir>..\\Analysis Code\\analcode.elf", pcfg, NULL);
	if (!pprog)
	{
		fprintf(stderr, "Cannot open %s\n", szapp);
		atck_endsession(psesn);
		return(1);
	}

	/*
	 * 5. Add instrumentation calls and/or perform
	 * static analysis
	 */
	pinitproto = atck_analproto(pprog, "void Initialize(uint32)");
	pcountproto = atck_analproto(pprog, "void CountProc()");
	atck_callbefore(pprog, pinitproto, 0);
	 
	/*
	 * 6. Step through code using iterators
	 */

	/* Step through each image in the program */
	pimgit = atck_imgit_new(pprog);
	pimg = atck_imgit_first(pimgit);
	while (pimg)
	{
		/* Step through each procedure in the image */
		pprocit = atck_img_procit_new(pimg);
		pproc = atck_procit_first(pprocit);
		while (pproc)
		{
			/*
			 * 7. Add instrumentation calls and/or perform
			 * static analysis
			 */
			if (atck_proc_attr(pproc) == ATCK_ATTR_INSTR)
			{
				/* Add instrmentation call */
				atck_proc_callbefore(pproc, pcountproto);
			}
			
			/* Perform static analysis */
			nprocs = nprocs + 1;
			
			/* Iterate to next procedure */
			pproc = atck_procit_next(pprocit);
		}
		atck_procit_free(pprocit);
		
		/* Write out each image as we finish with it */
		atck_img_write(pimg);

		/* Iterate to next image */
		pimg = atck_imgit_next(pimgit);
	}
	atck_imgit_free(pimgit);

	/*
	 * 8. Write out or close the application.
	 */
	piprog = atck_finish_write(pprog);
	
	/*
	 * 9. Download and run instrumented application.
	 */
	 
	if (brun)
	{
		/* Connect to the target system. */
		pdev = atck_connect(psesn, pcfg, HandleEvt, NULL);
		if (!pdev)
		{
			fprintf(stderr, "Cannot connect to target system\n");
			atck_endsession(psesn);
			return(1);
		}

		/* Download the instrumented application and start it running. */
		pcounter = atck_anal_symaddr(piprog, "nProcCount");
		prun = atck_idownload(pdev, piprog, NULL, NULL);
		atck_continue(prun);

		/* Stop the application whenever the user presses a key. */
		printf("Hit any key to stop program and get data\n");
		getchar();
		atck_stop(prun);
		
		/*
		 * 9. Analysis code gathers data in instrumented application.
		 *
		 * NOTE: SEE THE ANALYSIS CODE PROJECT FOR MORE DETAILS!
		 *
		 */

		/*
		 * 10. Read data from instrumented application.
		 *
		 * NOTE: THIS HAPPENS IN THE HandleEvt() FUNCTION!
		 *
		 */
	}
	
	/*
	 * 11. Analyze and output results
	 *
	 * NOTE: ALSO SEE THE HandleEvt() FUNCTION!
	 *
	 */
	printf("%s contains %d procedures.\n", szapp, nprocs);

	/*
	 * 12. Close ATtaCK
	 */
	if (brun)
		atck_disconnect(pdev);
	atck_endsession(psesn);
	return(0);
}


static void HandleEvt(atck_tevt_t*	pevt, void*, void*)
{
	atck_run_t* 	prun;
	atck_uint32_t	count;

	switch (atck_tevt_type(pevt))
	{
		case ATCK_TEVT_STOPPED:
			/* The application has stopped, so read and report data. */
			prun = atck_tevt_run(pevt);

			/*
			 * 10. Read data from instrumented application.
			 *
			 * NOTE: THIS HAPPENS IN THE HandleEvt() FUNCTION!
			 *
			 */
			atck_readdata(prun, pcounter, &count, "uint32", 1);
			
			/*
			 * 11. Analyze and output results
			 *
			 * NOTE: ALSO SEE THE HandleEvt() FUNCTION!
			 *
			 */
			printf("Procedures were called %d times.\n", count);

			break;

		case ATCK_TEVT_PRINT:
			/* The target wants to print some data to the host. */
			
			printf("%s", atck_tevt_printstr(pevt));
			break;

		case ATCK_TEVT_LOADPCT:
			/* These events show the progress of the download operation. */
			printf("Downloading... %u%%\n", atck_tevt_loadpct(pevt));
			break;
	}
}
